"""
This module is used as an interface to call the COPT solver for solving
(mixed-integer) linear or second-order cone programs.
"""

import coptpy as cp
import numpy as np
import warnings
import time
from scipy.sparse import csc_matrix
from .socp import SOCProg
from .lp import Solution


def solve(formula, display=True, params={}):

    try:
        if formula.xmat:
            warnings.warn('The SOCP solver ignores exponential cone constraints. ')
    except AttributeError:
        pass

    env = cp.Envr()
    m = env.createModel()
    m.setParam(cp.COPT.Param.Logging, False)
    m.setParam(cp.COPT.Param.LogToConsole, False)

    c = formula.obj
    A = formula.linear
    vtype = formula.vtype
    lhs = np.array([-cp.COPT.INFINITY]*formula.linear.shape[0])
    index_eq = formula.sense == 1
    lhs[index_eq] = formula.const[index_eq]
    rhs = formula.const

    lb = formula.lb
    ub = formula.ub
    lb[vtype == 'B'] = 0
    ub[vtype == 'B'] = 1

    m.loadMatrix(c, csc_matrix(A), lhs, rhs, lb, ub, vtype)

    if isinstance(formula, SOCProg):
        sc_dim = []
        sc_indices = []
        for q in formula.qmat:
            sc_dim.append(len(q))
            sc_indices.extend(q)

        ncone = len(formula.qmat)
        if ncone > 0:
            m.loadCone(ncone, None, sc_dim, sc_indices)

    if display:
        print('', 'Being solved by COPT...', sep='', flush=True)
        time.sleep(0.2)
    m.solve()
    stime = m.getAttr(cp.COPT.attr.SolvingTime)
    if all(vtype == 'C'):
        status = m.getAttr(cp.COPT.attr.LpStatus)
    else:
        status = m.getAttr(cp.COPT.attr.MipStatus)
    if display:
        print('Solution status: {0}'.format(status))
        print('Running time: {0:0.4f}s'.format(stime))

    try:
        x_sol = np.array(m.getValues())
        objval = formula.obj @ x_sol
        solution = Solution('COPT', objval, x_sol, status, stime)
    except cp.CoptError:
        warnings.warn('Fail to find the optimal solution.')
        # solution = None
        solution = Solution('COPT', np.nan, None, status, stime)

    return solution



"""
This module is used as an interface to call the Gurobi solver for solving
(mixed-integer) linear or second-order cone programs.
"""

from gurobipy import GurobiError
import gurobipy as gp
import numpy as np
import warnings
import time
from .socp import SOCProg
from .lp import Solution


def solve(formula, display=True, params={}):

    try:
        if formula.xmat:
            warnings.warn('The SOCP solver ignores exponential cone constraints. ')
        if formula.lmi:
            warnings.warn('The SOCP solver ignores semidefinite cone constraints. ')
    except AttributeError:
        pass

    nv = formula.linear.shape[1]
    vtype = list(formula.vtype)

    grb = gp.Model()
    x = grb.addMVar(nv, lb=formula.lb, ub=formula.ub, vtype=vtype)

    indices_eq = (formula.sense == 1)
    indices_ineq = (formula.sense == 0)
    linear_eq = formula.linear[indices_eq, :]
    linear_ineq = formula.linear[indices_ineq, :]
    const_eq = formula.const[indices_eq]
    const_ineq = formula.const[indices_ineq]
    if len(indices_eq) > 0:
        c_eq = grb.addMConstr(linear_eq, x, '=', const_eq)
        # grb.addMConstrs(linear_eq, x, '=', const_eq)
    if len(indices_ineq) > 0:
        c_ineq = grb.addMConstr(linear_ineq, x, '<', const_ineq)
        # grb.addMConstrs(linear_ineq, x, '<', const_ineq)

    if isinstance(formula, SOCProg):
        for constr in formula.qmat:
            index_right = constr[0:1]
            index_left = constr[1:]
            A = np.eye(len(index_left))
            grb.addConstr(x[index_left] @ A @ x[index_left] <=
                          x[index_right] @ x[index_right])

    grb.setObjective(formula.obj @ x)

    grb.setParam('LogToConsole', 0)
    try:
        for param, value in params.items():
            if eval('grb.Params.{}'.format(param)) is None:
                raise ValueError('Unknown parameter')
            grb.setParam(param, value)

    except (TypeError, ValueError):
        raise ValueError('Incorrect parameters or values.')
    if display:
        print('Being solved by Gurobi...', flush=True)
        time.sleep(0.2)
    grb.optimize()
    if display:
        print('Solution status: {0}'.format(grb.Status))
        print('Running time: {0:0.4f}s'.format(grb.Runtime))

    try:
        pi = np.ones(formula.linear.shape[0]) * np.nan
        upi = np.zeros(nv)
        lpi = np.zeros(nv)
        pi[indices_eq] = c_eq.pi
        pi[indices_ineq] = c_ineq.pi
        rc = x.rc
        upi[rc < 0] = rc[rc < 0]
        lpi[rc > 0] = rc[rc > 0]
        y = {'pi': pi, 'upi': upi, 'lpi': lpi}
    except GurobiError:
        y = None

    try:
        solution = Solution('Gurobi', grb.ObjVal, np.array(grb.getAttr('X')),
                            grb.Status, grb.Runtime, y=y)
    except AttributeError:
        warnings.warn('Fail to find the optimal solution.')
        # solution = None
        solution = Solution('Gurobi', np.nan, None, grb.Status, grb.Runtime)

    return solution



import numpy as np
import scipy.sparse as sp
from numbers import Real
from scipy.sparse import csr_matrix
from collections.abc import Iterable


def flat(a_list):
    flat_list = []
    for item in a_list:
        if isinstance(item, Iterable):
            flat_list.extend(flat(item))
        else:
            flat_list.append(item)

    return flat_list


def sparse_mul(ndarray, affine):

    array_ones = np.ones(ndarray.shape, dtype='int')
    affine_ind = np.arange(affine.size, dtype='int').reshape(affine.shape)
    affine_ones = np.ones(affine.shape, dtype='int')

    index = (array_ones * affine_ind).flatten()
    values = (ndarray * affine_ones).flatten()

    size = index.size

    return csr_matrix((values, (np.arange(size), index)),
                      shape=(size, affine.size))


def diag_comb(upper, lower):

    data = np.concatenate((upper.data, lower.data))
    indices = np.concatenate((upper.indices, upper.shape[1] + lower.indices))
    indptr = np.concatenate((upper.indptr[:-1], upper.indptr[-1] + lower.indptr))

    return csr_matrix((data, indices, indptr),
                      shape=np.array(upper.shape)+np.array(lower.shape))


def vert_comb(upper, lower):

    data = np.concatenate((upper.data, lower.data))
    indices = np.concatenate((upper.indices, lower.indices))
    indptr = np.concatenate((upper.indptr[:-1], upper.indptr[-1] + lower.indptr))

    shape = upper.shape[0] + lower.shape[0], max([upper.shape[1], lower.shape[1]])
    return csr_matrix((data, indices, indptr), shape=shape)


def sp_matmul(ndarray, affine, shape):

    size = int(np.prod(shape))

    if len(shape) < 1:
        return csr_matrix(ndarray)
    else:
        if len(affine.shape) == 1:
            affine = affine.reshape((affine.size, 1))
            row, col = shape[-1], 1
        elif len(ndarray.shape) == 1:
            ndarray = ndarray.reshape((1, ndarray.size))
            row, col = 1, shape[-1]
        else:
            row, col = shape[-2], shape[-1]

        inner = ndarray.shape[-1]

        affine_index = np.arange(affine.size).reshape(affine.shape)
        dim = len(affine.shape)
        axes = list(range(dim-2)) + [dim-1, dim-2]
        index = np.transpose(np.tile(affine_index, row), axes=axes).flatten()
        index_rep = size // (len(index)//inner)
        if index_rep > 1:
            index = np.tile(index, index_rep)

        data = np.tile(ndarray, col).flatten()
        data_rep = size // (len(data)//inner)
        if data_rep > 1:
            data = np.tile(data, data_rep)

        indptr = [inner*i for i in range(size+1)]

        return csr_matrix((data, index, indptr), shape=[size, affine.size])


def sp_lmatmul(ndarray, affine, shape):

    size = int(np.prod(shape))

    if len(shape) <= 0:
        return csr_matrix(ndarray)
    else:
        if len(affine.shape) == 1:
            affine = affine.reshape((1, affine.size))
            row, col = 1, shape[-1]
        elif len(ndarray.shape) == 1:
            ndarray = ndarray.reshape((ndarray.size, 1))
            row, col = shape[-1], 1
        else:
            row, col = shape[-2], shape[-1]

        inner = affine.shape[-1]

        affine_index = np.arange(affine.size).reshape(affine.shape)
        index = np.tile(affine_index, col).flatten()
        index_rep = size // (len(index)//inner)
        if index_rep > 1:
            index = np.tile(index, index_rep)

        dim = len(ndarray.shape)
        axes = list(range(dim-2)) + [dim-1, dim-2]
        data = np.transpose(np.tile(ndarray, row), axes=axes).flatten()
        data_rep = size // (len(data)//inner)
        if data_rep > 1:
            data = np.tile(data, data_rep)

        indptr = [inner*i for i in range(size+1)]

        return csr_matrix((data, index, indptr), shape=[size, affine.size])


def sp_trans(affine):

    index = np.arange(affine.size).reshape(affine.shape).T.flatten()
    indptr = np.arange(affine.size + 1)
    data = np.ones(affine.size)

    return csr_matrix((data, index, indptr), shape=[affine.size, affine.size])


def index_array(shape):

    if isinstance(shape, tuple):
        size = np.prod(shape)
    else:
        size = int(shape)
        shape = int(shape),

    return np.arange(size, dtype=int).reshape(shape)


def array_to_sparse(array):

    size = array.size
    all_items = list(array.reshape((size, )))

    indices = np.concatenate(tuple(item.indices for item in all_items))
    data = np.concatenate(tuple([item.data for item in all_items]))
    indices = indices[data != 0]
    data = data[data != 0]

    indptr = np.zeros(size+1)
    for i in range(size):
        indptr[1+i] = indptr[i] + all_items[i].count_nonzero()

    return csr_matrix((data, indices, indptr), (size, all_items[0].shape[1]))


def sv_to_csr(array):

    if not isinstance(array, np.ndarray):
        size = 1
        all_items = [array]
    else:
        size = array.size
        all_items = list(array.reshape((size, )))

    indices = np.concatenate(tuple(item.index for item in all_items))
    data = np.concatenate(tuple([item.value for item in all_items]))
    nonzero = (data != 0)
    indices = indices[nonzero]
    data = data[nonzero]

    indptr = np.zeros(size+1)
    for i in range(size):
        zero_counts = all_items[i].value.count(0)
        indptr[1+i] = indptr[i] + len(all_items[i].value) - zero_counts

    return csr_matrix((data, indices, indptr), (size, all_items[0].nvar))


def check_numeric(array):

    array = np.array(array.todense()) if sp.issparse(array) else array
    # array = np.array([array]) if not isinstance(array, np.ndarray) else array

    if isinstance(array, np.ndarray):
        if not isinstance(array.flat[0], np.number):
            raise TypeError('Incorrect data type of arrays.')
    elif not isinstance(array, Real):
        raise TypeError('Incorrect data type of arrays.')

    return array


def rso_broadcast(*args):

    arrays = [np.array(arg) if isinstance(arg, (Real, np.ndarray)) else arg.to_affine()
              for arg in args]

    indices = [np.arange(array.size).reshape(array.shape) for
               array in arrays]

    arrays = [array.reshape(array.size) for array in arrays]

    bdc = np.broadcast(*indices)
    outputs = [[item[i] for item, i in zip(arrays, index)] for index in bdc]

    return outputs


def add_linear(left, right):

    if left.shape[1] > right.shape[1]:
        right.resize((right.shape[0], left.shape[1]))
    elif right.shape[1] > left.shape[1]:
        left.resize((left.shape[0], right.shape[1]))

    return left + right


def event_dict(event_set):

    output = {}
    count = 0
    for item in event_set:
        for element in item:
            output[element] = count

        count += 1

    return output


def comb_set(s1, s2):

    d1 = event_dict(s1)
    d2 = event_dict(s2)

    dc = {item: str(d1[item]) + '-' + str(d2[item])
          for item in range(len(d1))}

    values = []
    output = []
    for key in dc:
        if dc[key] in values:
            index = values.index(dc[key])
            output[index].append(key)
        else:
            output.append([key])
            values.append(dc[key])

    return output


def E(expr):
    """
    The notion of the expectation of random variables and the worst-case
    expected values

    Notes
    -----
    This function is used to denote
    1. the expected value of an random variable when specifying the
    uncertainty set of expectations.
    2. the worst-case expected value of an RSOME expression involving random
    variables.
    """

    return expr.E



from .gcp import Model as GCPModel
from .gcp import GCProg
from .lp import LinConstr, Bounds, CvxConstr, ExpConstr
from .lp import Vars, VarSub, Affine, Convex
from .lp import def_sol
from .socp import Model as SOCModel
from .socp import SOCProg
from .subroutines import index_array, vert_comb
from collections.abc import Iterable
from numbers import Real
from scipy.sparse import csr_matrix

import numpy as np
import pandas as pd
import scipy.sparse as sp


class Model:

    def __init__(self, name=None):

        self.master = GCPModel(nobj=False, mtype='R', name=name, top=self)
        self.here_now_index = []
        self.wait_see_index = []
        self.all_here_now_index = []
        self.rand_index = []

        self.subs = []

        self.exact_bounds = []
        self.relaxed_bounds = []
        self.solutions = []

    def dvar(self, shape=(), vtype='C', name=None):

        if self.subs:
            msg = 'No here-and-now decisions can be defined '
            msg += 'after the definition of subproblems.'
            raise RuntimeError(msg)

        new_var = self.master.dvar(shape, vtype, name, aux=False)
        index = range(new_var.first, new_var.last)
        self.here_now_index.extend(index)

        return HaNVar(new_var)

    def rvar(self, shape=(), name=None):

        new_var = self.master.dvar(shape, 'C', name, aux=False)
        index = range(new_var.first, new_var.last)
        self.rand_index.extend(index)

        return RandVar(new_var)

    def min(self, obj):

        if isinstance(obj, (Vars, VarSub, Affine)):
            affine = obj.to_affine()
            if affine.linear[:, self.wait_see_index].nnz > 0:
                msg = 'No wait-and-see decision is allowed in master problem objective.'
                raise ValueError(msg)
            if affine.linear[:, self.rand_index].nnz > 0:
                msg = 'No random variable is allowed in master problem objective.'
                raise ValueError(msg)
        elif isinstance(obj, Convex):
            cond1 = obj.affine_in.linear[:, self.wait_see_index].nnz > 0
            cond2 = obj.affine_out.linear[:, self.wait_see_index].nnz > 0
            if cond1 or cond2:
                msg = 'No wait-and-see decision is allowed in master problem objective.'
                raise ValueError(msg)
            cond1 = obj.affine_in.linear[:, self.rand_index].nnz > 0
            cond2 = obj.affine_out.linear[:, self.rand_index].nnz > 0
            if cond1 or cond2:
                msg = 'No random variable is allowed in master problem objective.'
                raise ValueError(msg)
        else:
            raise TypeError('Unsupported objective function.')

        self.master.min(obj)

    def max(self, obj):

        if isinstance(obj, (Vars, VarSub, Affine)):
            affine = obj.to_affine()
            if affine.linear[:, self.wait_see_index].nnz > 0:
                msg = 'No wait-and-see decision is allowed in master problem objective.'
                raise ValueError(msg)
            if affine.linear[:, self.rand_index].nnz > 0:
                msg = 'No random variable is allowed in master problem objective.'
                raise ValueError(msg)
        elif isinstance(obj, Convex):
            cond1 = obj.affine_in.linear[:, self.wait_see_index].nnz > 0
            cond2 = obj.affine_out.linear[:, self.wait_see_index].nnz > 0
            if cond1 or cond2:
                msg = 'No wait-and-see decision is allowed in master problem objective.'
                raise ValueError(msg)
            cond1 = obj.affine_in.linear[:, self.rand_index].nnz > 0
            cond2 = obj.affine_out.linear[:, self.rand_index].nnz > 0
            if cond1 or cond2:
                msg = 'No random variable is allowed in master problem objective.'
                raise ValueError(msg)
        else:
            raise TypeError('Unsupported objective function.')

        self.master.max(obj)

    def st(self, *arg):

        for constr in arg:
            if isinstance(constr, Iterable):
                for item in constr:
                    self.st(item)
            else:
                if isinstance(constr, LinConstr):
                    if constr.linear[:, self.wait_see_index].nnz > 0:
                        msg = 'No wait-and-see decision is allowed in master problem.'
                        raise ValueError(msg)
                    if constr.linear[:, self.rand_index].nnz > 0:
                        msg = 'No random variable is allowed in master problem.'
                        raise ValueError(msg)

                elif isinstance(constr, CvxConstr):
                    cond1 = constr.affine_in.linear[:, self.wait_see_index].nnz > 0
                    cond2 = constr.affine_out.linear[:, self.wait_see_index].nnz > 0
                    if cond1 or cond2:
                        msg = 'No wait-and-see decision is allowed in master problem.'
                        raise ValueError(msg)
                    cond1 = constr.affine_in.linear[:, self.rand_index].nnz > 0
                    cond2 = constr.affine_out.linear[:, self.rand_index].nnz > 0
                    if cond1 or cond2:
                        msg = 'No random variable is allowed in master problem.'
                        raise ValueError(msg)

                elif isinstance(constr, Bounds):
                    if set(constr.indices).intersection(self.wait_see_index):
                        msg = 'No wait-and-see decision is allowed in master problem.'
                        raise ValueError(msg)
                    if set(constr.indices).intersection(self.rand_index):
                        msg = 'No random variable is allowed in master problem.'
                        raise ValueError(msg)

                elif isinstance(constr, ExpConstr):
                    print('TODO!')
                    raise ValueError('Not yet.')

                else:
                    raise TypeError('Unsupported constraints.')

                self.master.st(constr)

    def subprob(self, size=1, lower=-1e10, upper=1e10):

        sub_models = SubModel(self, size)
        new_var = self.master.dvar(size)
        sub_models.values = new_var
        sub_models.value_index = list(range(new_var.first, new_var.last))
        self.st(new_var >= lower, new_var <= upper)

        self.subs.append(sub_models)

        return sub_models

    def do_math(self):

        formula = self.master.do_math()
        all_indices = np.arange(0, formula.linear.shape[1])
        indices = np.setxor1d(all_indices,
                              np.array(self.wait_see_index + self.rand_index),
                              assume_unique=True)

        formula = GCProg(formula.linear[:, indices], formula.const, formula.sense,
                         formula.vtype[indices],
                         formula.ub[indices], formula.lb[indices],
                         formula.qmat, formula.xmat, formula.obj[:, indices])

        self.all_here_now_index = indices

        return formula

    def solve(self, solver=None, display=True, params={}):

        self.master.solve(solver, display, params)
        # self.relaxed_bounds.append(self.master.get())
        self.solutions.append(self.master.solution)

    def add_cut(self, cuts):

        add_linear = sp.vstack([cut.linear for cut in cuts])
        add_const = np.concatenate([cut.const for cut in cuts])
        add_sense = np.zeros(len(cuts))

        formula = self.master.primal
        formula.linear = sp.vstack((formula.linear, add_linear), format='csr')
        formula.const = np.concatenate((formula.const, add_const))
        formula.sense = np.concatenate((formula.sense, add_sense))
        self.master.primal = formula

    def update(self, solver=None, master_display=True, sub_display=False,
               master_params={}, sub_params={}):

        self.solve(solver, master_display, master_params)
        self.relaxed_bounds.append(self.master.get())
        obj_linear = self.master.obj.linear.toarray()[0]
        obj_const = self.master.obj.const
        indices = self.here_now_index
        x_current = np.array(self.master.solution.x)[indices]
        exact_obj = obj_linear[self.here_now_index] @ x_current + obj_const

        for sub in self.subs:
            sol = sub.solve(solver, sub_display, sub_params)
            objvals = sol['objval']
            exact_obj += obj_linear[sub.value_index] @ np.array(objvals)
            self.add_cut(sol['cut'])

        self.exact_bounds.append(exact_obj.reshape(()))

    def get_bounds(self):

        if self.master.sign > 0:
            lower = max(self.relaxed_bounds)
            upper = min(self.exact_bounds)
        else:
            lower = max(self.exact_bounds)
            upper = min(self.relaxed_bounds)

        return lower, upper

    def get_gap(self, relative=True):

        lower, upper = self.get_bounds()

        if relative:
            return (upper - lower) / lower
        else:
            return upper - lower

    def get(self):

        if self.exact_bounds == []:
            raise RuntimeError('The model is unsolved or no solution is obtained.')
        else:
            if self.master.sign > 0:
                return min(self.exact_bounds)
            else:
                return max(self.exact_bounds)


class SubModel:

    def __init__(self, model, size):

        self.model = model
        self.obj = None
        self.sign = None
        self.size = size
        self.lin_constr = []
        self.bounds = []
        self.values = None
        self.value_index = None
        self.wait_see_index = []
        self.uset = [None] * size
        self.z_center = {i: np.ones(len(self.model.rand_index)) * np.nan
                         for i in range(size)}
        self.z_uncertain = {i: [] for i in range(size)}

        rand_index = model.rand_index
        self.rand_index_map = pd.Series(range(len(rand_index)), index=rand_index)

        self.formula = None
        self.update = True

    def __getitem__(self, item):

        return SubModelIter(self, item)

    def def_obj(self, obj, sign):

        if isinstance(obj, (Vars, VarSub, Affine)):
            affine = obj.to_affine()
            if affine.linear[:, self.model.here_now_index].nnz > 0:
                msg = 'Here-and-now decisions are not allowed in sub-problem objective.'
                raise ValueError(msg)
            if affine.linear[:, self.model.rand_index].nnz > 0:
                msg = 'Random variables are not allowed in sub-problem objective.'
                raise ValueError(msg)
            value_index = list(range(self.values.first, self.values.last))
            if affine.linear[:, value_index].nnz > 0:
                msg = 'Sub-problem values are not allowed in sub-problem objective.'
                raise ValueError(msg)
        else:
            raise TypeError('Unsupported objective function')

        self.obj = obj
        self.sign = sign

    def dvar(self, shape=(), name=None):

        new_var = self.model.master.dvar(shape, 'C', name, aux=False)
        index = range(new_var.first, new_var.last)
        self.model.wait_see_index.extend(index)
        self.wait_see_index.extend(index)

        self.update = True

        return new_var

    def min(self, obj):

        self.def_obj(obj, 1)

        self.update = True

    def max(self, obj):

        self.def_obj(obj, -1)

        self.update = True

    def st(self, *arg):

        for constr in arg:
            if isinstance(constr, Iterable):
                for item in constr:
                    self.st(item)
            else:
                value_index = list(range(self.values.first, self.values.last))
                if isinstance(constr, LinConstr):
                    if constr.linear[:, value_index].nnz > 0:
                        msg = 'Sub-problem values are not allowed in sub-problem.'
                        raise ValueError(msg)
                    if constr.linear[:, self.model.wait_see_index].nnz == 0:
                        msg = 'No wait-and-see decision appears in the constraint.'
                        raise ValueError(msg)
                    self.lin_constr.append(constr)
                elif isinstance(constr, Bounds):
                    if not set(constr.indices).intersection(self.model.wait_see_index):
                        msg = 'No wait-and-see decision appears in the constraint.'
                        raise ValueError(msg)
                    self.bounds.append(constr)
                else:
                    raise TypeError('Unsupported constraints.')

        self.update = True

    def uncertain(self, *args):

        self[:].uncertain(*args)

    def mean(self, weights=None):

        values = self.values
        num_scen = values.shape[0]
        if weights is None:
            weights = np.ones(num_scen) / num_scen
        else:
            weights = weights / weights.sum()

        return values @ weights

    def sum(self, weights=None):

        values = self.values
        num_scen = values.shape[0]
        if weights is None:
            weights = np.ones(num_scen)

        return values @ weights

    def do_math(self):

        if self.update is False and self.formula is not None:
            return self.formula

        sub_model = SOCModel(nobj=True, mtype='R')

        vars = []
        for var in self.model.master.vars:
            if var.vtype == 'C':
                vars.append(var)
            else:
                new_var = Vars(var.model, var.first, var.shape, 'C',
                               var.name, var.sparray)
                vars.append(new_var)
        sub_model.vars = vars + self.model.master.auxs
        sub_model.lin_constr = self.lin_constr
        sub_model.bounds = self.bounds

        dual = sub_model.do_math(primal=False, obj=False)
        dual.const = self.obj.linear.toarray()[0] * dual.const * self.sign

        vtype = dual.vtype
        ub = dual.ub
        lb = dual.lb
        qmat = dual.qmat
        obj = dual.obj

        linear = dual.linear[self.wait_see_index, :]
        const = dual.const[self.wait_see_index]
        sense = dual.sense[self.wait_see_index]

        cmat = dual.linear[self.model.rand_index, :].T
        if self.model.all_here_now_index == []:
            all_indices = np.arange(0, dual.linear.shape[0])
            not_here_now = self.model.wait_see_index + self.model.rand_index
            indices = np.setxor1d(all_indices, np.array(not_here_now),
                                  assume_unique=True)
            self.model.all_here_now_index = indices
        else:
            indices = self.model.all_here_now_index

        amat = dual.linear[indices, :].T

        # num_rand = len(self.model.rand_index)
        dprog = {}
        for i in range(self.size):
            z_uncertain = self.z_uncertain[i]

            milm = SOCModel(nobj=True)
            lmb = milm.dvar(cmat.shape[0])
            tvec = lmb @ cmat
            bigm = 1e6
            obj_index = []
            for uset in z_uncertain:
                r = uset['radius']
                idx = uset['index']
                if uset['utype'] == '1':
                    t = milm.dvar()
                    obj_index.extend(t.get_ind())
                    u = milm.dvar(len(idx), vtype='B')
                    v = milm.dvar(len(idx), vtype='B')

                    milm.st(t <= r*tvec[idx] + (1-u)*bigm)
                    milm.st(t <= -r*tvec[idx] + (1-v)*bigm)
                    milm.st((u + v).sum() == 1)
                elif uset['utype'] == 'I':
                    t = milm.dvar(len(idx))
                    obj_index.extend(t.get_ind())
                    u = milm.dvar(len(idx), vtype='B')
                    v = milm.dvar(len(idx), vtype='B')

                    milm.st(t <= r*tvec[idx] + (1-u)*bigm)
                    milm.st(t <= -r*tvec[idx] + (1-v)*bigm)
                    milm.st((u + v) == 1)

            dp_formula = milm.do_math()
            dp_formula.obj = np.zeros(dp_formula.obj.shape)
            dp_formula.obj[0, obj_index] = 1
            dprog[i] = dp_formula

        formula = SubProg(linear, const, sense, vtype, ub, lb,
                          qmat, amat, cmat, dprog, obj)
        self.formula = formula
        self.update = False

        return formula

    def solve(self, solver=None, display=False, params={}):

        formula = self.do_math()

        indices = self.model.all_here_now_index
        x_here_now = np.array(self.model.master.solution.x)[indices]

        cuts = []
        pis = []
        objvals = []

        for i, z_values in self.z_center.items():
            obj = formula.obj + formula.amat@x_here_now
            obj += formula.cmat@z_values
            obj = obj.reshape((1, obj.size))

            if not self.z_uncertain[i]:
                linear = formula.linear
                const = formula.const
                sense = formula.sense
                vtype = formula.vtype
                ub, lb = formula.ub, formula.lb
                qmat = formula.qmat
            else:
                dprog = formula.dprog[i]
                lmb_dim = obj.shape[1]

                linear = vert_comb(formula.linear, dprog.linear)
                const = np.concatenate((formula.const, dprog.const))
                sense = np.concatenate((formula.sense, dprog.sense))
                vtype = np.concatenate((formula.vtype, dprog.vtype[lmb_dim:]))
                ub = np.concatenate((formula.ub, dprog.ub[lmb_dim:]))
                lb = np.concatenate((formula.lb, dprog.lb[lmb_dim:]))
                qmat = formula.qmat
                obj = np.concatenate((obj, -dprog.obj[:, lmb_dim:]), axis=1)

            sub_formula = SOCProg(linear, const, sense, vtype, ub, lb, qmat, obj)

            if solver is None:
                sub_sol = def_sol(sub_formula, display, params)
            else:
                sub_sol = solver.solve(sub_formula, display, params)

            pi = sub_sol.x[:formula.amat.shape[0]]
            pis.append(pi)
            objvals.append(- (sub_sol.objval) * self.sign)

            cut_linear = - (pi @ formula.amat)
            cut_linear = csr_matrix(cut_linear.reshape((1, cut_linear.size)))
            cut_const = (formula.obj - formula.cmat@z_values) @ pi
            if self.z_uncertain[i]:
                xi = sub_sol.x[formula.amat.shape[0]:]
                cut_const -= xi @ obj[0, formula.amat.shape[0]:]
            # print(xi)
            left = Affine(self.model.master, cut_linear, cut_const)
            right = self.values[i]
            cut = left <= self.sign * right
            cuts.append(cut)

        return {'pi': pis, 'objval': objvals, 'cut': cuts}


class SubModelIter:

    def __init__(self, submodel, item):

        self.submodel = submodel
        self.item = item

    def uncertain(self, *args):

        for uncertainty in args:

            if isinstance(uncertainty, Uncertainty):
                if isinstance(self.item, int):
                    self.item = slice(self.item, self.item+1)
                if not isinstance(self.item, slice):
                    raise TypeError('Unsupported indexes.')

                index = self.submodel.rand_index_map.loc[uncertainty.vars.get_ind()]
                for i in range(self.submodel.size)[self.item]:
                    z_center = self.submodel.z_center[i]

                    if not all(np.isnan(z_center[index])):
                        msg = 'Uncertainty of the same random variable '
                        msg += 'cannot be redefined.'
                        raise ValueError(msg)
                    z_center[index] = uncertainty.center.flatten()
                    self.submodel.z_center[i] = z_center

                    if uncertainty.utype in ['1', 'I']:
                        temp = {'utype': uncertainty.utype,
                                'index': index.values,
                                'radius': uncertainty.radius}
                        self.submodel.z_uncertain[i].append(temp)
            else:
                raise TypeError('Unsupported uncertainty set.')


class SubProg(SOCProg):

    def __init__(self, linear, const, sense, vtype, ub, lb,
                 qmat, amat, cmat, dprog, obj):

        super().__init__(linear, const, sense, vtype, ub, lb, qmat, obj)
        self.amat = amat
        self.cmat = cmat
        self.dprog = dprog


class RandVar(Vars):

    def __init__(self, vars):

        super().__init__(vars.model, vars.first, vars.shape, vars.vtype,
                         vars.name, vars.sparray)

    def __repr__(self):

        string = super().__repr__()

        return string.replace('continuous ', '').replace('decision', 'random')

    def __getitem__(self, item):

        item_array = index_array(self.shape)
        indices = item_array[item]

        return RandVarSub(self, indices)

    def singleton(self, center):

        if isinstance(center, Real):
            center = np.array([center])

        center = center.reshape(self.shape)

        return Uncertainty(self, center, 0, utype='S')

    def l1_ball(self, center, radius):

        if isinstance(center, Real):
            center = np.array([center])

        center = center.reshape(self.shape)

        return Uncertainty(self, center, radius, utype='1')

    def box(self, center, radius):

        if isinstance(center, Real):
            center = np.array([center])

        center = center.reshape(self.shape)

        return Uncertainty(self, center, radius, utype='I')

    def get(self):

        raise TypeError('Cannot access the solution of random variables')


class RandVarSub(VarSub):

    def __init__(self, var, indices):

        super().__init__(var, indices)

    def singleton(self, center):

        if isinstance(center, Real):
            center = np.array([center])

        center = center.reshape(self.indices.shape)

        return Uncertainty(self, center, 0, utype='S')

    def l1_ball(self, center, radius):

        if isinstance(center, Real):
            center = np.array([center])

        center = center.reshape(self.indices.shape)

        return Uncertainty(self, center, radius, utype='1')

    def box(self, center, radius):

        if isinstance(center, Real):
            center = np.array([center])

        center = center.reshape(self.indices.shape)

        return Uncertainty(self, center, radius, utype='I')

    def get(self):

        raise TypeError('Cannot access the solution of random variables')


class HaNVar(Vars):

    def __init__(self, vars):

        super().__init__(vars.model, vars.first, vars.shape, vars.vtype,
                         vars.name, vars.sparray)

    def get(self):

        top = self.model.top
        exact_bounds = np.array(top.exact_bounds)
        index = np.argmin(top.master.sign * exact_bounds)

        xvec = top.solutions[index].x[self.first:self.last]

        return xvec.reshape(self.shape)


class Uncertainty:

    def __init__(self, vars, center, radius, utype):

        self.vars = vars
        self.center = center
        self.radius = radius
        self.utype = utype

    def __repr__(self):

        utypes = {'S': 'singleton',
                  '1': 'one-norm',
                  'I': 'infinite-norm'}

        article = 'an' if self.utype == 'I' else 'a'
        string = f'{article} {utypes[self.utype].lower()} uncertainty set '
        string += f'for {self.vars.__repr__()}'

        return string



from .math import norm, square, sumsqr, quad, rsocone
from .math import expcone, exp, log, pexp, plog, entropy, kldiv
from .math import maxof, minof
from .math import trace, diag, tril, triu
from .subroutines import E
from .lp import concat, rstack, cstack



from .lp import PiecewiseConvex
from .lp import Vars, VarSub, Affine
from .lp import DecRule, DecRuleSub
from .lp import RoAffine
from .subroutines import *
from numbers import Real
import numpy as np


def norm(affine, degree=2):
    """
    Return the first, second, or infinity norm of a 1-D array.

    Parameters
    ----------
    affine : an array of variables or affine expressions.
        Input array. The array must be 1-D.
    degree : {1, 2, numpy.inf}, optional
        Order of the norm function. It can only be 1, 2, or infinity.
        The default value is 2, i.e., the Euclidean norm.

    Returns
    -------
    out : Convex
        The norm of the given array.
    """

    return affine.norm(degree)


def square(affine):
    """
    Return the element-wise square of an array.

    Parameters
    ----------
    affine : an array of variables or affine expression
        Input array.

    Returns
    -------
    out : Convex
        The element-wise squares of the given array
    """

    return affine.to_affine().square()


def sumsqr(affine):
    """
    Return the sum of squares of a 1-D array.

    Parameters
    ----------
    affine : an array of variables or affine expression
        Input array. The array must be 1-D.

    Returns
    -------
    out : Convex
        The sum of squares of the given array
    """

    return affine.to_affine().sumsqr()


def quad(x, qmat):
    """
    Return the quadratic expression x @ qmat @ x.

    Parameters
    ----------
    x : an array of variables or affine expressions
        Input array. The array must be 1-D.
    qmat : a positive or negative semidefinite matrix.

    Returns
    -------
    out : Convex
        The quadratic expression affine @ qmat affine
    """

    return x.to_affine().quad(qmat)


def rsocone(x, y, z):
    """
    Return a rotated cone constraint sumsqr(x) <= y*z.

    Parameters
    ----------
    x : an array of variables or affine expressions
        Input array. The array must be a vector.
    y : {Real, Vars, VarSub, Affine}
        The y value in the constraint. It must be a scalar.
    z : {Real, Vars, VarSub, Affine}
        The z value in the constraint. It must be a scalar.

    Returns
    -------
    output : CvxConstr
        The rotated cone constraint
    """

    return x.to_affine().rsocone(y, z)


def exp(affine):
    """
    Return the element-wise natural exponential function
    exp(affine).

    Parameters
    ----------
    affine : an array of variables or affine expression
        Input array.

    Returns
    -------
    out : Convex
        The natural exponential function exp(affine)
    """

    return affine.exp()


def log(affine):
    """
    Return the element-wise natural logarithm function
    log(affine).

    Parameters
    ----------
    affine : an array of variables or affine expression
        Input array.

    Returns
    -------
    out : Convex
        The natural logarithm function log(affine)
    """

    return affine.log()


def pexp(affine, scale):
    """
    Return the element-wise perspective of natural exponential
    function scale * exp(affine/scale).

    Parameters
    ----------
    affine : an array of variables or affine expression
        Input array.

    scale : {Real, Vars, VarSub, Affine}
        The scale value in the perspective function. It must be a
        scalar.

    Returns
    -------
    out : PerspConvex
        The perspective of natural exponential function
        scale * exp(affine/scale)
    """

    return affine.pexp(scale)


def plog(affine, scale):
    """
    Return the element-wise perspective of natural logarithm
    function scale * log(affine/scale).

    Parameters
    ----------
    affine : an array of variables or affine expression
        Input array.

    scale : {Real, Vars, VarSub, Affine}
        The scale value in the perspective function. It must be a
        scalar.

    Returns
    -------
    out : PerspConvex
        The perspective of natural logarithm function
        scale * log(affine/scale)
    """

    return affine.plog(scale)


def entropy(affine):
    """
    Return the entropy expression sum(affine*log(affine)).

    Parameters
    ----------
    affine : an array of variables or affine expression
        Input array. It must be a vector.

    Returns
    -------
    out : Convex
        The entropy expression sum(affine*log(affine))
    """

    return affine.entropy()


def expcone(y, x, z):
    """
    Return an exponential cone constraint z*exp(x/z) <= y.

    Parameters
    ----------
    y : an array of variables or affine expression
        The right-hand-side of the constraint
    x : {Real, Vars, VarSub, Affine}
        The x value in the constraint. It must be a scalar.
    z : {Real, Vars, VarSub, Affine}
        The z value in the constraint. It must be a scalar.

    Returns
    -------
    constr : ExpConstr
        The exponential cone constraint z*exp(x/z) <= y
    """

    return y.expcone(x, z)


def kldiv(p, q, r):
    """
    Return an KL divergence constraint sum(p*log(p/phat)) <= r.

    Parameters
    ----------
    p : an array of variables or affine expression
        The array of probabilities. It must be a vector.
    q : {Real, Vars, VarSub, Affine}
        The array of empirical probabilities. It must a vector with
        the same shape as p.
    r : {Real, Vars, VarSub, Affine}
        The ambiguity constant. It must be a scalar.

    Returns
    -------
    constr : ExpConstr
        The KL divergence constraint sum(p*log(p/q)) <= r.
    """

    return p.kldiv(q, r)


def trace(affine):
    """
    Return the trace of a 2-D array.

    Parameters
    ----------
    affine : an array of variables or affine expressions.
        Input array. It must be a 2-D array.

    Returns
    -------
    out : Affine
        Output array as the trace of the 2-D array.
    """

    return affine.trace()


def maxof(*args):
    """
    Return a piecewise function of the maximum of a number of
    expressions.

    Parameters
    ----------
    *args : RSOME affine or bi-affine expressions, real numbers
        Expressions or numerical values used to define the piecewise
        function.

    Returns
    -------
    piecewise : PiecewiseConvex
        The piecewise function defined to be the maximum of a number
        of given expressions or numerical values.
    """

    pieces = flat(args)
    this_model = None
    for arg in pieces:
        if isinstance(arg, (Vars, VarSub, Affine)):
            arg = arg.to_affine()
            if this_model is None:
                this_model = arg.model.top
            elif arg.model.top is not this_model:
                raise ValueError('Models not match.')
        elif isinstance(arg, RoAffine):
            if this_model is None:
                this_model = arg.affine.model.top
        elif isinstance(arg, (DecRule, DecRuleSub)):
            arg = arg.to_affine()
            if this_model is None:
                this_model = arg.model.top
        elif isinstance(arg, Real):
            arg = np.array([arg])
        elif not isinstance(arg, np.ndarray):
            raise TypeError('Unsupported expressions.')

        if arg.size != 1:
            raise ValueError('All pieces must have their sizes to be one.')

    if this_model is None:
        raise ValueError('Incorrect piecewise function.')

    return PiecewiseConvex(this_model, pieces, sign=1, add_sign=1)


def minof(*args):
    """
    Return a piecewise function of the minimum of a number of
    expressions.

    Parameters
    ----------
    *args : RSOME affine or bi-affine expressions, real numbers
        Expressions or numerical values used to define the piecewise
        function.

    Returns
    -------
    piecewise : PiecewiseConvex
        The piecewise function defined to be the minimum of a number
        of given expressions or numerical values.
    """

    piecewise = maxof(*args)
    piecewise.pieces = [-piece for piece in piecewise.pieces]

    piecewise.sign = -1
    piecewise.add_sign = -1

    return piecewise


def diag(affine, k=0, fill=False):
    """
    Return the diagonal elements of a 2-D array.

    Parameters
    ----------
    affine : an array of variables or affine expressions.
        Input array. It must be a 2-D array.

    k : int, optional
        Diagonal in question. The default is 0. Use `k>0` for diagonals
        above the main diagonal, and `k<0` for diagonals below the main
        diagonal.

    fill : bool
        If True, return a 2-D array where the non-diagonal elements are
        filled with zeros. Otherwise, return the diagonal elements as a
        1-D array

        Returns
        -------
        out : Affine
            The diagonal elements of a given 2-D array.
    """

    return affine.diag(k, fill)


def tril(affine, k=0):
    """
    Return the lower triangular elements of a 2-D array. The remaining
    elements are filled with zeros.

        Parameters
        ----------
        affine : an array of variables or affine expressions.
            Input array. It must be a 2-D array.

        k : int, optional
            Diagonal above which to zero elements.  `k = 0` (the
            default) is the main diagonal, `k < 0` is below it and
            `k > 0` is above.

        Returns
        -------
        out : Affine
            The lower triangular elements of the given 2-D array.
    """

    return affine.tril(k)


def triu(affine, k):
    """
    Return the upper triangular elements of a 2-D array. The remaining
    elements are filled with zeros.

    Parameters
    ----------
    affine : an array of variables or affine expressions.
        Input array. It must be a 2-D array.

    k : int, optional
        Diagonal above which to zero elements.  `k = 0` (the default)
        is the main diagonal, `k < 0` is below it and `k > 0` is above.

    Returns
    -------
        out : Affine
        The upper triangular elements of the given 2-D array.
    """

    return affine.triu(k)



praise = """Praise of DRO

In the world of optimization,
where models reign supreme,
there is one approach that stands out,
a star that shines and gleams.

Distributionally robust optimization,
a name both grand and bold,
is the method that ensures our models,
will not be easily controlled.

For in the real world, distributions vary,
and models can easily fail,
but DRO is a safeguard,
against this common tale.

It guards against worst-case scenarios,
and makes our models tough,
so they can handle any data,
and perform when things get rough.

DRO, we praise your robustness,
your ability to withstand,
the challenges of the real world,
and the tests that it demands.

You bring stability to our models,
and give us confidence in our results,
so we can make decisions with assurance,
and never fear the unknowns.

So here's to you, distributionally robust optimization,
our trusted friend and guide,
may you continue to protect our models,
and keep them safe inside.
"""

print(praise)



from .lp import Affine, CvxConstr, PCvxConstr, ExpConstr, KLConstr, LMIConstr
from .lp import def_sol, Solution
from .socp import Model as SOCModel
from .socp import SOCProg
import numpy as np
import pandas as pd
import scipy.sparse as sp
from collections.abc import Iterable
from .subroutines import flat, rso_broadcast, vert_comb


class Model(SOCModel):
    """
    The Model class creates an SOCP model object
    """

    def __init__(self, nobj=False, mtype='R', name=None, top=None):

        super().__init__(nobj, mtype, name, top)
        self.psd_vars = []
        self.exp_constr = []
        self.other_constr = []

    def reset(self):

        self.lin_constr = []
        self.pws_constr = []
        self.cone_constr = []
        self.exp_constr = []
        self.other_constr = []
        self.bounds = []
        self.aux_constr = []
        self.aux_bounds = []
        self.cvx_constr = []

    def st(self, constr):

        if isinstance(constr, Iterable):
            for item in constr:
                self.st(item)
        elif isinstance(constr, ExpConstr):
            self.exp_constr.append(constr)
        elif isinstance(constr, CvxConstr):
            if constr.xtype in 'XLP':
                self.other_constr.append(constr)
            else:
                super().st(constr)
        elif isinstance(constr, KLConstr):
            self.other_constr.append(constr)
        elif isinstance(constr, LMIConstr):
            self.other_constr.append(constr)
            expr = Affine(constr.model, constr.linear, constr.const)
            super().st(expr == expr.T)
        else:
            super().st(constr)

        self.pupdate = True
        self.dupdate = True

        return constr

    def do_math(self, primal=True, refresh=True, obj=True):
        """
        Return the linear or conic programming problem as the
        standard formula of the model.

        Parameters
        ----------
        primal : bool, default True
            Specify whether return the primal formula of the model.
            If primal=False, the method returns the daul formula.

        refresh : bool
            Leave the argument unspecified.

        obj : bool
            Leave the argument unspecified.

        Returns
        -------
        prog : GCProg
            A conic programming problem.
        """

        if primal:
            if self.primal is not None and not self.pupdate:
                return self.primal

            if refresh:
                self.auxs = []
                self.aux_constr = []
                self.aux_bounds = []
                self.last = self.vars[-1].first + self.vars[-1].size

            more_exp = []
            if self.obj is not None:
                obj_constr = (self.vars[0] - self.sign * self.obj >= 0)
                if isinstance(obj_constr, CvxConstr):
                    constr = obj_constr
                    if constr.xtype in 'XLP':
                        self.other_constr.append(constr)

            lmi = []
            for constr in self.other_constr:
                if isinstance(constr, KLConstr):
                    ns = constr.p.size
                    aux_var = self.dvar(ns, aux=True)
                    self.aux_constr.append(aux_var.sum() <= constr.r)
                    for s in range(ns):
                        ps = constr.phat[s]
                        p = constr.p[s]

                        entro_constr = aux_var[s] * (1/ps) >= -(p * (1/ps)).entropy()
                        exp_constr = ExpConstr(entro_constr.model,
                                               entro_constr.affine_out,
                                               1, entro_constr.affine_in)

                        more_exp.append(exp_constr)
                elif isinstance(constr, PCvxConstr):
                    if constr.xtype == 'X':
                        affine_out = constr.affine_out * (1/constr.multiplier)
                        exprs_list = rso_broadcast(constr.affine_in,
                                                   constr.affine_scale,
                                                   affine_out)
                        for exprs in exprs_list:
                            exp_cone_constr = ExpConstr(constr.model,
                                                        exprs[0],
                                                        -exprs[2], exprs[1])
                            self.exp_constr.append(exp_cone_constr)
                    elif constr.xtype == 'L':
                        affine_out = constr.affine_out * (1/constr.multiplier)
                        exprs_list = rso_broadcast(constr.affine_in,
                                                   constr.affine_scale,
                                                   affine_out)
                        for exprs in exprs_list:
                            exp_cone_constr = ExpConstr(constr.model,
                                                        exprs[2], exprs[0], exprs[1])
                            self.exp_constr.append(exp_cone_constr)
                elif isinstance(constr, CvxConstr):
                    if constr.xtype == 'P':
                        affine_out = constr.affine_out * (1/constr.multiplier)
                        aux_var = self.dvar(constr.affine_in.shape)
                        self.aux_constr.append(aux_var.sum() >= affine_out)
                        ns = constr.affine_in.size
                        affine_in = constr.affine_in.reshape(ns)
                        affine_aux = aux_var.to_affine().reshape(ns)
                        for s in range(ns):
                            exp_cone_constr = ExpConstr(constr.model,
                                                        affine_aux[s],
                                                        1, affine_in[s])
                            more_exp.append(exp_cone_constr)
                    elif constr.xtype == 'X':
                        affine_out = constr.affine_out * (1/constr.multiplier)
                        exprs_list = rso_broadcast(constr.affine_in, affine_out)
                        for exprs in exprs_list:
                            exp_cone_constr = ExpConstr(constr.model,
                                                        exprs[0], -exprs[1], 1)
                            self.exp_constr.append(exp_cone_constr)
                    elif constr.xtype == 'L':
                        affine_out = constr.affine_out * (1/constr.multiplier)
                        exprs_list = rso_broadcast(constr.affine_in, affine_out)
                        for exprs in exprs_list:
                            exp_cone_constr = ExpConstr(constr.model,
                                                        exprs[1], exprs[0], 1)
                            self.exp_constr.append(exp_cone_constr)
                elif isinstance(constr, LMIConstr):
                    lmi.append({'linear': constr.linear,
                                'const': constr.const,
                                'dim': constr.dim})

            xmat = []
            for constr in self.exp_constr + more_exp:
                aux_var = self.dvar(3, aux=True)
                self.aux_constr.append(aux_var[0] - constr.expr1 == 0)
                self.aux_constr.append(aux_var[1] - constr.expr2 <= 0)
                self.aux_constr.append(aux_var[2] - constr.expr3 == 0)
                xmat.append(list(range(aux_var.first, aux_var.first + 3)))

            formula = super().do_math(primal=True, refresh=False, obj=obj)
            formula = GCProg(formula.linear, formula.const, formula.sense,
                             formula.vtype, formula.ub, formula.lb,
                             formula.qmat, xmat, lmi, formula.obj)
            self.primal = formula
            self.pupdate = False

            return formula

        else:
            if self.dual is not None and not self.dupdate:
                return self.dual

            primal = self.do_math(obj=obj)
            pvar_num = primal.linear.shape[1]

            dual_socp = super().do_math(primal=False, refresh=False, obj=obj)

            if len(primal.xmat) == 0 and len(primal.lmi) == 0:
                formula = GCProg(dual_socp.linear, dual_socp.const, dual_socp.sense,
                                 dual_socp.vtype, dual_socp.ub, dual_socp.lb,
                                 dual_socp.qmat, [], [], dual_socp.obj)
                self.dual = formula
                return formula

            if len(primal.qmat) == 0:
                pxmat = primal.xmat
                plmi = primal.lmi
            else:
                num_exp = len(primal.xmat)
                i_idx = np.array([range(num_exp)] * 3).T.flatten()
                j_idx = flat(primal.xmat)
                sp_xmat = sp.csr_matrix(([1, 1, 1] * num_exp,
                                         (i_idx, j_idx)),
                                        (num_exp, pvar_num))
                socp_idx = flat(primal.qmat)
                keep_idx = [i for i in range(pvar_num)
                            if i not in socp_idx]
                sp_xmat = sp_xmat[:, keep_idx]
                pxmat = [list(sp_xmat[i].indices) for i in range(num_exp)]

                for each in primal.lmi:
                    if each['linear'].shape[1] < pvar_num:
                        each['linear'].resize((each['linear'].shape[0], pvar_num))
                    each['linear'] = each['linear'][:, keep_idx]
                plmi = primal.lmi

            linear = dual_socp.linear
            const = dual_socp.const
            sense = dual_socp.sense
            obj = dual_socp.obj
            vtype = dual_socp.vtype
            ub = dual_socp.ub
            lb = dual_socp.lb
            qmat = dual_socp.qmat

            num_xc = len(pxmat)
            if num_xc > 0:
                count_col = np.arange(0, 3*num_xc, 3).reshape((num_xc, 1))
                xmat = count_col + np.array([[0, 1, 2]]*num_xc)
                xmat = list((linear.shape[1] + xmat))
                data = [-1, 1, -1, -1] * num_xc
                i_idx = np.array([ex + ex[-1:] for ex in pxmat]).flatten()
                j_idx = (count_col + np.array([2, 1, 0, 2])).flatten()
                extra_block = sp.csr_matrix((data, (i_idx, j_idx)),
                                            (linear.shape[0], 3*num_xc))

                linear = sp.hstack((linear, extra_block))
                obj = np.hstack((obj, np.zeros(3*num_xc)))
                ub = np.hstack((ub, np.inf*np.ones(3*num_xc)))
                lb = np.hstack((lb, -np.inf*np.ones(3*num_xc)))
                vtype = np.hstack((vtype, np.array(['C']*3*num_xc)))
            else:
                xmat = []

            num_lmi = len(plmi)
            if num_lmi > 0:
                linear_list = []
                const_list = [obj]
                dims = []
                lmi = []
                here = 0
                total_col = linear.shape[1] + sum([each['dim']**2 for each in plmi])
                for each in plmi:
                    each_linear = each['linear']
                    if each_linear.shape[1] != linear.shape[0]:
                        each_linear.resize([each_linear.shape[0], linear.shape[0]])
                    # elif each_linear.shape[1] > linear.shape[0]:
                    #     linear.resize([each_linear.shape[1], linear.shape[1]])
                    linear_list.append(each_linear)

                    each_const = -each['const']
                    const_list.append(each_const.flatten())

                    each_dim = each['dim']
                    dims.append(each['dim'])

                    each_row = each_dim ** 2
                    each_idx = np.arange(each_row)
                    temp_idx = linear.shape[1] + here + each_idx
                    lmi_linear = sp.csr_matrix((np.ones(each_row),
                                                (each_idx, temp_idx)),
                                               (each_row, total_col))
                    here += each_row
                    lmi.append({'linear': lmi_linear,
                                'const': np.zeros((each_dim, each_dim)),
                                'dim': each_dim})

                extra_block = sp.csr_matrix(sp.vstack(linear_list).T)

                linear = sp.hstack((linear, extra_block))
                obj = np.hstack((const_list))
                extra_dim = np.square(dims).sum()
                ub = np.hstack((ub, np.inf*np.ones(extra_dim)))
                lb = np.hstack((lb, -np.inf*np.ones(extra_dim)))
                vtype = np.hstack((vtype, np.array(['C']*extra_dim)))
            else:
                lmi = []

            formula = GCProg(sp.csr_matrix(linear), const, sense,
                             vtype, ub, lb, qmat, xmat, lmi, obj)

            self.dual = formula
            self.dupdate = False
            return formula

    def soc_solve(self, solver=None, degree=4, cuts=(-30, 60),
                  display=True, params={}):
        """
        Solve the approximated SOC model with the selected solver interface.

        Parameters
        ----------
            solver : {None, lpg_solver, clp_solver, ort_solver, eco_solver
                      cpx_solver, grb_solver, msk_solver}
                Solver interface used for model solution. Use default solver
                if solver=None.
            degree : int
                The L-degree value for approximating exponential cone
                constraints.
            cuts : tuple of two integers
                The lower and upper cut-off values for the SOC approximation
                of exponential constraints.
            display : bool
                Display option of the solver interface.
            params : dict
                A dictionary that specifies parameters of the selected solver.
                So far the argument only applies to Gurobi, CPLEX, and Mosek.
        """

        formula = self.do_math(obj=True).to_socp(degree, cuts)
        if solver is None:
            solution = def_sol(formula, display, params)
        else:
            solution = solver.solve(formula, display, params)

        if isinstance(solution, Solution):
            self.solution = solution
        else:
            self.solution = None


class GCProg(SOCProg):
    """
    The GCProg class creates an exponential/semidefinite cone program
    """

    def __init__(self, linear, const, sense, vtype, ub, lb, qmat, xmat, lmi, obj=None):

        super().__init__(linear, const, sense, vtype, ub, lb, qmat, obj)
        self.xmat = xmat
        self.lmi = lmi

    def __repr__(self):

        xmat = self.xmat
        lmi = self.lmi
        string = 'Conic program object:\n'
        socp_str = super().__repr__()
        string += socp_str[socp_str.find('\n')+1:]
        string += '---------------------------------------------\n'
        string += 'Number of ExpCone constraints: {0}\n'.format(len(xmat))
        string += '---------------------------------------------\n'
        string += 'Number of PSCone constraints:  {0}\n'.format(len(lmi))

        return string

    def showec(self):

        n = len(self.xmat)

        if n == 0:
            return None

        indices = np.concatenate([item for item in self.xmat])
        values = [1, 2, 3] * n
        indptr = range(0, 3*n+1, 3)

        var_names = ['x{0}'.format(i)
                     for i in range(1, self.linear.shape[1] + 1)]
        constr_names = ['EC{0}'.format(j)
                        for j in range(1, n + 1)]
        table = pd.DataFrame(sp.csr_matrix((values, indices, indptr),
                             (n, self.linear.shape[1])).todense(),
                             index=constr_names, columns=var_names)
        table['sense'] = ['-'] * n
        table['constant'] = ['-'] * n

        return table

    def showlmi(self):

        n = len(self.lmi)
        if n == 0:
            return None

        linear_list = []
        num_col = self.linear.shape[1]
        for c in self.lmi:
            each_linear = c['linear']
            if each_linear.shape[1] < num_col:
                each_linear.resize((each_linear.shape[0], num_col))
            linear_list.append(each_linear)

        left = sp.vstack(linear_list).toarray()

        right = np.concatenate([c['const'].flatten() for c in self.lmi])

        var_names = ['x{0}'.format(i)
                     for i in range(1, self.linear.shape[1] + 1)]

        constr_names = np.concatenate([[f'PSDC{i+1}'] * (c['dim'] ** 2)
                                       for i, c in enumerate(self.lmi)])

        table = pd.DataFrame(left, columns=var_names, index=constr_names)

        table['sense'] = '>>'
        table['constant'] = right

        return table

    def show(self):
        """
        Returns a pandas.DataFrame that summarizes the information on the
        optimization problem.
        """

        table = self.showlc()
        obj_row = pd.DataFrame(self.obj.reshape((1, self.obj.size)),
                               columns=table.columns[:-2], index=['Obj'])
        table = pd.concat([obj_row, table], axis=0)

        table_qc = self.showqc()
        if table_qc is not None:
            table = pd.concat([table, table_qc], axis=0)
        table_ec = self.showec()
        if table_ec is not None:
            table = pd.concat([table, table_ec], axis=0)
        table_lmi = self.showlmi()
        if table_lmi is not None:
            table = pd.concat([table, table_lmi], axis=0)

        ub = pd.DataFrame(self.ub.reshape((1, self.ub.size)),
                          columns=table.columns[:-2], index=['UB'])
        lb = pd.DataFrame(self.lb.reshape((1, self.lb.size)),
                          columns=table.columns[:-2], index=['LB'])
        vtype = pd.DataFrame(self.vtype.reshape((1, self.vtype.size)),
                             columns=table.columns[:-2], index=['Type'])
        table = pd.concat([table, ub, lb, vtype], axis=0)

        return table.fillna('-')

    def to_socp(self, degree=4, cuts=(-30, 60)):

        num_vars = 1 + 4 + degree + 3
        num_cols = num_vars + (3 + degree)*3

        t_idx = 0
        x_idx = [1, 2]
        alpha_idx = [3, 4]
        fgh_idx = [5, 6, 7]
        v_idx = list(range(8, 8 + degree))

        data = [1]
        row_idx = [0]
        col_idx = [0]
        row_count = 1

        data += [1]*4
        row_idx += [1, 1, 2, 2]
        col_idx += x_idx + alpha_idx
        row_count += 2

        data += [20/2**degree/24, 23/24, 0.25, 1/24, -1]
        row_idx += [row_count]*5
        col_idx += [x_idx[1], alpha_idx[1], fgh_idx[0], fgh_idx[2], v_idx[0]]
        row_count += 1

        cut_lower, cut_upper = cuts
        data += [1, -cut_lower, 1, -cut_upper, -1, cut_lower]
        row_idx += list(np.array([0, 0, 1, 1, 2, 2]) + row_count)
        col_idx += [x_idx[0], alpha_idx[0], x_idx[1],
                    alpha_idx[1], x_idx[1], alpha_idx[1]]
        row_count += 3

        here = num_vars
        ones = [-1, -1, 1]
        data += [0.5, -0.5, 1/2**degree, -0.5, -0.5] + ones
        row_idx += list(np.array([0, 0, 1, 2, 2] + list(range(3))) + row_count)
        col_idx += [alpha_idx[1], fgh_idx[0], x_idx[1],
                    alpha_idx[1], fgh_idx[0]] + list(range(here, here+3))
        here += 3
        row_count += 3

        data += [0.5, -0.5, 1/2**degree, 1, -0.5, -0.5] + ones
        row_idx += list(np.array([0, 0, 1, 1, 2, 2] + list(range(3))) + row_count)
        col_idx += [alpha_idx[1], fgh_idx[1], x_idx[1], alpha_idx[1],
                    alpha_idx[1], fgh_idx[1]] + list(range(here, here+3))
        here += 3
        row_count += 3

        data += [0.5, -0.5, 1, -0.5, -0.5] + ones
        row_idx += list(np.array([0, 0, 1, 2, 2] + list(range(3))) + row_count)
        col_idx += [alpha_idx[1], fgh_idx[2], fgh_idx[1],
                    alpha_idx[1], fgh_idx[2]] + list(range(here, here+3))
        here += 3
        row_count += 3

        for d in range(degree - 1):
            data += [0.5, -0.5, 1, -0.5, -0.5] + ones
            row_idx += list(np.array([0, 0, 1, 2, 2] + list(range(3))) + row_count)
            col_idx += [alpha_idx[1], v_idx[d+1], v_idx[d],
                        alpha_idx[1], v_idx[d+1]] + list(range(here, here+3))
            here += 3
            row_count += 3

        data += [0.5, -0.5, 1, -0.5, -0.5] + ones
        row_idx += list(np.array([0, 0, 1, 2, 2] + list(range(3))) + row_count)
        col_idx += [alpha_idx[1], t_idx, v_idx[-1],
                    alpha_idx[1], t_idx] + list(range(here, here+3))
        here += 3
        row_count += 3

        more_linear = sp.csr_matrix((data, (row_idx, col_idx)),
                                    shape=(row_count, num_cols))
        more_sense = [0] + [1]*2 + [0]*4 + [1, 1, 0]*(3 + degree)
        more_const = np.zeros(row_count)
        more_lb = - np.ones(num_cols) * np.inf
        more_lb[alpha_idx + fgh_idx + v_idx] = 0
        for d in range(3 + degree):
            more_lb[num_vars + d*3 + 2] = 0

        left_width = self.linear.shape[1]
        qmat = self.qmat

        linear = self.linear
        const = self.const
        sense = self.sense
        ub = self.ub
        lb = self.lb
        obj = self.obj
        vtype = self.vtype
        right_width = more_linear.shape[1]
        for xm in self.xmat:
            left_width = linear.shape[1]
            # print(left_width)
            left_linear = sp.csr_matrix((-np.ones(3),
                                         (range(3), [xm[1], xm[0], xm[2]])),
                                        shape=(row_count, left_width))
            linear = vert_comb(linear, sp.hstack((left_linear, more_linear)))
            # print(linear.shape)
            const = np.concatenate((const, more_const))
            sense = np.concatenate((sense, more_sense))
            ub = np.concatenate((ub, np.ones(right_width) * np.inf))
            lb = np.concatenate((lb, more_lb))
            obj = np.concatenate((obj, np.zeros(right_width)))
            vtype = np.concatenate((vtype, np.array(['C']*right_width)))

            qmat += [list(left_width + num_vars + np.array([2, 1, 0]) + q*3)
                     for q in range(3+degree)]

        return SOCProg(linear, const, sense, vtype, ub, lb, qmat, obj)



from .gcp import Model as GCPModel
from .ro import Model as ROModel
from .lp import DecBounds, DecExpConstr, LinConstr
from .lp import ConeConstr, PCvxConstr, CvxConstr, ExpConstr, LMIConstr
from .lp import Vars, Affine
from .lp import RoAffine, RoConstr
from .lp import DecVar, RandVar, DecLinConstr, DecCvxConstr, DecPCvxConstr
from .lp import DecRoConstr
from .lp import PiecewiseConvex, PWConstr, ExpPWConstr, DecLMIConstr
from .lp import Scen
from .lp import Solution, def_sol
from .subroutines import event_dict
import numpy as np
import pandas as pd
import scipy.sparse as sp
from numbers import Real
from collections.abc import Sized, Iterable


class Model:
    """
    Returns a model object with the given number of scenarios.

    Parameters
    ----------
    scens : int or array-like objects
        The number of scenarios, if it is an integer. It could also be
        an array of scenario indices.
    name : str
        Name of the model

    Returns
    -------
    model : rsome.dro.Model
        A model object
    """

    def __init__(self, scens=1, name=None):

        self.ro_model = ROModel()
        self.vt_model = GCPModel(mtype='V', top=self)
        self.sup_model = self.ro_model.sup_model
        self.sup_model.top = self
        self.ro_model.rc_model.top = self
        self.exp_model = GCPModel(nobj=True, mtype='E', top=self)
        self.pro_model = GCPModel(nobj=True, mtype='P', top=self)

        self.obj_ambiguity = None

        if isinstance(scens, int):
            num_scen = scens
            series = pd.Series(np.arange(num_scen).astype(int))
        elif isinstance(scens, Sized):
            num_scen = len(scens)
            series = pd.Series(np.arange(num_scen).astype(int), index=list(scens))
        else:
            raise TypeError('Incorrect scenario type.')
        self.num_scen = num_scen
        self.series_scen = series

        self.dec_vars = [DecVar(self, self.vt_model.vars[0])]
        self.rand_vars = []
        self.all_constr = []
        self.var_ev_list = None
        # self.affadapt_mat = None

        pr = self.pro_model.dvar(num_scen, name='probabilities')
        self.p = pr

        self.obj = None
        self.sign = 1

        self.primal = None
        self.dual = None
        self.solution = None
        self.pupdate = True
        self.dupdate = True

        self.solution = None

        self.name = name

    def rvar(self, shape=(), name=None):
        """
        Returns an array of random variables with the given shape.

        Parameters
        ----------
        shape : int or tuple
            Shape of the variable array.
        name : str
            Name of the variable array

        Returns
        -------
        new_var : rsome.lp.RandVar
            An array of new random variables
        """

        sup_var = self.sup_model.dvar(shape, 'C', name)
        exp_var = self.exp_model.dvar(shape, 'C', name)
        rand_var = RandVar(sup_var, exp_var)
        self.rand_vars.append(rand_var)

        return rand_var

    def dvar(self, shape=(), vtype='C', name=None):
        """
        Returns an array of decision variables with the given shape
        and variable type.

        Parameters
        ----------
        shape : int or tuple
            Shape of the variable array.
        vtype : {'C', 'B', 'I'}
            Type of the decision variables. 'C' means continuous; 'B'
            means binary, and 'I" means integer.
        name : str
            Name of the variable array

        Returns
        -------
        new_var : rsome.lp.DecVar
            An array of new decision variables
        """

        dec_var = self.vt_model.dvar(shape, vtype, name)
        dec_var = DecVar(self, dec_var, name=name)
        self.dec_vars.append(dec_var)

        return dec_var

    def ambiguity(self):
        """
        Returns an event-wise ambiguity set with the given number of scenarios
        decalred in creating the DRO model.
        """

        if self.all_constr:
            raise SyntaxError('Ambiguity set must be specified ' +
                              'before defining constraints.')

        return Ambiguity(self)

    def rule_var(self):

        if self.var_ev_list is not None:
            return self.var_ev_list

        total = sum(dvar.size*len(dvar.event_adapt)
                    for dvar in self.dec_vars)
        vtype = ''.join([dvar.vtype * dvar.size * len(dvar.event_adapt)
                         if len(dvar.vtype) == 1
                         else dvar.vtype * len(dvar.event_adapt)
                         for dvar in self.dec_vars])
        var_const = self.ro_model.dvar(total, vtype=vtype)

        count = 0
        for dvar in self.dec_vars:
            dvar.ro_first = count
            count += dvar.size*len(dvar.event_adapt)

        num_scen = self.num_scen
        self.var_ev_list = []
        for s in range(num_scen):
            start = 0
            index = []
            total_size = 0
            total_col = 0
            for dvar in self.dec_vars:
                edict = event_dict(dvar.event_adapt)
                size = dvar.size
                index.extend(list(start + size * edict[s] +
                                  np.arange(size, dtype=int)))

                start += size * len(dvar.event_adapt)
                total_size += size
                total_col += size * len(dvar.event_adapt)

            self.var_ev_list.append(var_const[index].to_affine())

        if self.sup_model.vars:

            adapt_list = [dvar.rand_adapt if dvar.rand_adapt is not None else
                          np.zeros((dvar.size, self.sup_model.vars[-1].last))
                          for dvar in self.dec_vars]
            depend_mat = np.concatenate(adapt_list, axis=0)
            if depend_mat.sum() > 0:

                num_depend = np.array([int(dvar.rand_adapt.sum())
                                       if dvar.rand_adapt is not None else 0
                                       for dvar in self.dec_vars])
                scen_depend = np.array([len(dvar.event_adapt)
                                        for dvar in self.dec_vars])
                total_depend = num_depend @ scen_depend
                each_depend = num_depend.sum()
                var_linear = self.ro_model.dvar(total_depend)
                nz_rows = np.where(depend_mat.flatten())[0]
                for s in range(num_scen):
                    start = 0
                    index = []
                    total_var = 0
                    for dvar, num, scen in zip(self.dec_vars,
                                               num_depend, scen_depend):
                        if num == 0:
                            continue
                        edict = event_dict(dvar.event_adapt)
                        index.extend(list(start + num * edict[s] +
                                          np.arange(num, dtype=int)))
                        start += num * scen
                        total_var += num

                    depend_affine = var_linear[index].to_affine()

                    nz_cols = depend_affine.linear.indices
                    last = depend_affine.linear.shape[1]

                    ra_linear = sp.csr_matrix(([1.0] * each_depend,
                                               (nz_rows, nz_cols)),
                                              shape=(depend_mat.size, last))
                    ra_const = np.zeros(depend_mat.shape)
                    raffine = Affine(self.ro_model.rc_model,
                                     ra_linear, ra_const)
                    self.var_ev_list[s] = RoAffine(raffine,
                                                   self.var_ev_list[s],
                                                   self.ro_model.sup_model)

        return self.var_ev_list

    def min(self, obj):
        """
        Minimize the given objective function.

        Parameters
        ----------
        obj
            An objective function

        Notes
        -----
        The objective function given as an array must have the size
        to be one.
        """

        if self.obj is not None:
            raise SyntaxError('Redefinition of the objective is not allowed.')

        if not isinstance(obj, (Real, PiecewiseConvex)):
            if obj.size > 1:
                raise ValueError('Incorrect function dimension.')

        self.obj = obj
        self.sign = 1
        self.pupdate = True
        self.dupdate = True

    def max(self, obj):
        """
        Maximize the given objective function.

        Parameters
        ----------
        obj
            An objective function

        Notes
        -----
        The objective function given as an array must have the size
        to be one.
        """

        if self.obj is not None:
            raise SyntaxError('Redefinition of the objective is not allowed.')

        if not isinstance(obj, (Real, PiecewiseConvex)):
            if obj.size > 1:
                raise ValueError('Incorrect function dimension.')

        self.obj = obj
        self.sign = - 1
        self.pupdate = True
        self.dupdate = True

    def minsup(self, obj, ambset):
        """
        Minimize the worst-case expected objective value over the given
        ambiguity set.

        Parameters
        ----------
        obj
            Objective function involving random variables
        ambset : Ambiguity
            The ambiguity set defined for the worst-case expectation

        Notes
        -----
        The ambiguity set defined for the objective function is considered
        the default ambiguity set for the distritionally robust model.
        """

        if self.obj is not None:
            raise SyntaxError('Redefinition of the objective is not allowed.')

        if not isinstance(obj, (Real, PiecewiseConvex)):
            if obj.size > 1:
                raise ValueError('Incorrect function dimension.')

        self.obj = obj
        self.obj_ambiguity = ambset
        self.sign = 1
        self.pupdate = True
        self.dupdate = True

    def maxinf(self, obj, ambset):

        """
        Maximize the worst-case expected objective value over the given
        ambiguity set.

        Parameters
        ----------
        obj
            Objective function involving random variables
        ambset : Ambiguity
            The ambiguity set defined for the worst-case expectation

        Notes
        -----
        The ambiguity set defined for the objective function is considered
        the default ambiguity set for the distritionally robust model.
        """

        if self.obj is not None:
            raise SyntaxError('Redefinition of the objective is not allowed.')

        if not isinstance(obj, (Real, PiecewiseConvex)):
            if obj.size > 1:
                raise ValueError('Incorrect function dimension.')

        self.obj = obj
        self.obj_ambiguity = ambset
        self.sign = - 1
        self.pupdate = True
        self.dupdate = True

    def st(self, *arg):
        """
        Define constraints that an optimization model subject to.

        Parameters
        ----------
        *args
            Constraints or collections of constraints that the model
            subject to.

        Notes
        -----
        Multiple constraints can be defined altogether as the argument
        of the st method.
        """

        for constr in arg:
            if isinstance(constr, Iterable):
                for item in constr:
                    self.st(item)
            else:
                if isinstance(constr, (DecLinConstr, DecBounds,
                                       DecCvxConstr, DecPCvxConstr,
                                       DecExpConstr, DecLMIConstr)):
                    if constr.model is not self.vt_model:
                        raise ValueError('Models mismatch.')
                elif isinstance(constr, DecRoConstr):
                    if constr.dec_model is not self.vt_model or \
                       constr.rand_model is not self.sup_model:
                        raise ValueError('Models mismatch.')
                elif isinstance(constr, (PWConstr, ExpPWConstr)):
                    if constr.model is not self:
                        raise ValueError('Models mismatch.')
                else:
                    raise TypeError('Unsupported constraints.')

                if isinstance(constr, ExpPWConstr):
                    self.all_constr.append(constr)
                elif isinstance(constr, PWConstr):
                    self.all_constr.extend(constr.pieces)
                else:
                    self.all_constr.append(constr)

        self.pupdate = True
        self.dupdate = True

    def do_math(self, primal=True):
        """
        Return the linear, second-order cone, or exponential cone
        programming problem as the standard formula or robust
        counterpart of the model.

        Parameters
        ----------
        primal : bool, default True
            Specify whether return the primal formula of the model.
            If primal=False, the method returns the daul formula.

        Returns
        -------
        prog : GCProg
            An exponential cone programming problem.
        """

        if primal:
            if self.primal is not None and not self.pupdate:
                return self.primal
        else:
            if self.dual is not None and not self.dupdate:
                return self.dual
            else:
                self.do_math(primal=True)
                return self.ro_model.do_math(False)

        self.ro_model.reset()
        self.rule_var()

        # Event-wise objective function
        self.ro_model.obj = None
        self.ro_model.min(self.ro_model.rc_model.vars[1][0].to_affine())
        sign = self.sign
        constr = (self.dec_vars[0] >= self.obj * sign)
        if isinstance(constr, ExpPWConstr):
            ro_constr_list = self.dro_to_roc(constr)
        elif isinstance(constr, PWConstr):
            ro_constr_list = []
            for piece in constr.pieces:
                ro_constr_list.extend(self.ro_to_roc(piece))
        elif isinstance(constr, DecCvxConstr):
            ro_constr_list = self.ro_to_roc(constr)
        elif constr.ctype == 'R':
            ro_constr_list = self.ro_to_roc(constr)
        elif constr.ctype == 'E':
            ro_constr_list = self.dro_to_roc(constr)
        else:
            raise TypeError('Unsupported objective function.')

        self.ro_model.st(ro_constr_list)

        # Event-wise Constraints
        for constr in self.all_constr:
            if isinstance(constr, ExpPWConstr):
                ro_constr_list = self.dro_to_roc(constr)
            elif isinstance(constr, (DecCvxConstr, DecPCvxConstr, DecExpConstr)):
                ro_constr_list = self.ro_to_roc(constr)
            elif constr.ctype == 'R':
                ro_constr_list = self.ro_to_roc(constr)
            elif constr.ctype == 'E':
                ro_constr_list = self.dro_to_roc(constr)
            else:
                raise ValueError('Unknown constraints.')
            self.ro_model.st(ro_constr_list)

        formula = self.ro_model.do_math(primal)
        if primal:
            self.primal = formula
            self.pupdate = False
        else:
            self.dual = formula
            self.dupdate = False

        return formula

    def ro_to_roc(self, constr):

        drule_list = self.rule_var()
        num_var = self.vt_model.vars[-1].last

        ro_constr = []
        # for event in constr.event_adapt:
        for s in range(self.num_scen):
            drule = drule_list[s]
            if isinstance(constr, DecRoConstr):
                is_equal = (all(constr.sense) if
                            isinstance(constr.sense, Iterable) else
                            constr.sense == 1)
                if is_equal:
                    roaffine = RoAffine(constr.raffine, constr.affine,
                                        constr.rand_model)
                    left = DecRoConstr(roaffine, 0,
                                       constr.event_adapt, constr.ctype)
                    left.ambset = constr.ambset
                    right = DecRoConstr(-roaffine, 0,
                                        constr.event_adapt, constr.ctype)
                    right.ambset = constr.ambset

                    return self.ro_to_roc(left) + self.ro_to_roc(right)

                raf_linear, aff_linear = (constr.raffine.linear[:, :num_var],
                                          constr.affine.linear[:, :num_var])

                row_ind = np.unique(raf_linear.indices)
                if isinstance(drule, RoAffine):
                    if len(row_ind) > 0:
                        if (drule.raffine[row_ind].linear.nnz > 0 or
                                np.any(drule.raffine[row_ind].const)):
                            raise SyntaxError('Incorrect affine expressions.')

                    raffine = raf_linear @ drule.affine
                    raffine = (raffine.reshape(constr.raffine.shape) +
                               constr.raffine.const)
                    roaffine = RoAffine(raffine, np.zeros(aff_linear.shape[0]),
                                        self.sup_model)
                    this_const = constr.affine.const
                    this_const = this_const.flatten()
                    roaffine = roaffine + aff_linear@drule + this_const
                    ew_constr = RoConstr(roaffine, constr.sense)

                elif isinstance(drule, Affine):
                    raffine = raf_linear @ drule
                    raffine = (raffine.reshape(constr.raffine.shape) +
                               constr.raffine.const)
                    this_const = constr.affine.const
                    this_const = this_const.flatten()
                    roaffine = RoAffine(raffine, aff_linear@drule + this_const,
                                        self.sup_model)
                    ew_constr = RoConstr(roaffine, constr.sense)
                else:
                    raise TypeError('Unknown type.')

            elif isinstance(constr, DecLinConstr):
                linear = constr.linear
                const = constr.const
                roaffine = linear @ drule - const.reshape(const.size)
                if isinstance(roaffine, RoAffine):
                    is_equal = (all(constr.sense) if
                                isinstance(constr.sense, Iterable) else
                                constr.sense == 1)
                    if is_equal == 1:

                        left = DecLinConstr(constr.model,
                                            constr.linear, constr.const,
                                            np.zeros(constr.linear.shape[0]),
                                            constr.event_adapt, constr.ctype)
                        right = DecLinConstr(constr.model,
                                             -constr.linear, -constr.const,
                                             np.zeros(constr.linear.shape[0]),
                                             constr.event_adapt, constr.ctype)
                        return self.ro_to_roc(left) + self.ro_to_roc(right)

                    left_empty = roaffine.raffine.linear.nnz == 0
                    right_empty = not roaffine.raffine.const.any()
                    if left_empty and right_empty:
                        ew_constr = LinConstr(roaffine.dec_model,
                                              roaffine.affine.linear,
                                              - roaffine.affine.const,
                                              constr.sense)
                    else:
                        ew_constr = RoConstr(roaffine, constr.sense)
                elif isinstance(roaffine, Affine):
                    ew_constr = LinConstr(roaffine.model, roaffine.linear,
                                          -roaffine.const, constr.sense)
                else:
                    raise TypeError('Unknown type.')
            elif isinstance(constr, DecPCvxConstr):
                linear_in = constr.affine_in.linear
                const_in = constr.affine_in.const
                aff_in = linear_in@drule + const_in.reshape(const_in.size)
                aff_in = aff_in.reshape(constr.affine_in.shape)

                if isinstance(constr.affine_scale, (Real, np.ndarray)):
                    aff_scale = constr.affine_scale
                else:
                    scale = constr.affine_scale.to_affine()
                    linear_sc = scale.linear
                    const_sc = scale.const
                    aff_scale = linear_sc@drule + const_sc.reshape(const_sc.size)
                aff_scale = aff_scale.reshape(constr.affine_scale.shape)

                if isinstance(aff_in, RoAffine):
                    aff_in = aff_in.affine
                if isinstance(constr.affine_out, (np.ndarray, Real)):
                    linear_out = np.zeros((constr.affine_out.size, drule.shape[0]))
                    const_out = constr.affine_out
                else:
                    linear_out = constr.affine_out.linear
                    const_out = constr.affine_out.const
                aff_out = linear_out@drule + const_out.reshape(const_out.size)
                if isinstance(aff_out, RoAffine):
                    aff_out = aff_out.affine
                aff_out = aff_out.reshape(constr.affine_out.shape)

                ew_constr = PCvxConstr(aff_in.model, aff_in, aff_scale, aff_out,
                                       constr.multiplier, constr.xtype)
            elif isinstance(constr, DecCvxConstr):
                linear_in = constr.affine_in.linear
                const_in = constr.affine_in.const
                aff_in = linear_in@drule + const_in.reshape(const_in.size)
                if isinstance(aff_in, RoAffine):
                    aff_in = aff_in.affine
                if isinstance(constr.affine_out, (np.ndarray, Real)):
                    linear_out = np.zeros((constr.affine_out.size, drule.shape[0]))
                    const_out = constr.affine_out
                else:
                    linear_out = constr.affine_out.linear
                    const_out = constr.affine_out.const
                aff_out = linear_out@drule + const_out.reshape(const_out.size)
                if isinstance(aff_out, RoAffine):
                    aff_out = aff_out.affine
                ew_constr = CvxConstr(aff_in.model, aff_in, aff_out,
                                      constr.multiplier, constr.xtype)
            elif isinstance(constr, DecExpConstr):
                if isinstance(drule, RoAffine):
                    drule_affine = drule.affine
                else:
                    drule_affine = drule

                affine1 = constr.expr1.to_affine()
                expr1 = affine1.linear@drule_affine + affine1.const

                if isinstance(constr.expr2, Real):
                    expr2 = constr.expr2
                else:
                    affine2 = constr.expr2.to_affine()
                    linear2 = affine2.linear
                    const2 = affine2.const
                    expr2 = linear2@drule_affine + const2

                if isinstance(constr.expr3, Real):
                    expr3 = constr.expr3
                else:
                    affine3 = constr.expr3.to_affine()
                    linear3 = affine3.linear
                    const3 = affine3.const
                    expr3 = linear3@drule_affine + const3

                ew_constr = ExpConstr(expr1.model, expr1, expr2, expr3)

            elif isinstance(constr, DecLMIConstr):
                if isinstance(drule, RoAffine):
                    drule_affine = drule.affine
                else:
                    drule_affine = drule

                lmi_left = constr.linear @ drule_affine - constr.const.flatten()
                lmi_linear = lmi_left.linear
                lmi_const = (-lmi_left.const).reshape((constr.dim, constr.dim))

                ew_constr = LMIConstr(lmi_left.model, lmi_linear, lmi_const, constr.dim)

            else:
                raise TypeError('Unknown constraint type.')

            if isinstance(ew_constr, RoConstr):
                if (ew_constr.raffine.linear.nnz > 0 or
                        np.any(ew_constr.raffine.const)):
                    if constr.ambset is None:
                        if self.obj_ambiguity is None:
                            raise SyntaxError('The Ambiguity set is '
                                              'undefined.')
                        else:
                            ambset = self.obj_ambiguity
                            support = ambset.sup_constr[s]
                    else:
                        ambset = constr.ambset
                        if isinstance(ambset, Ambiguity):
                            support = constr.ambset.sup_constr[s]
                        elif isinstance(ambset, Iterable):
                            support = ambset
                    ew_constr = ew_constr.forall(support)
                else:
                    ew_constr = LinConstr(ew_constr.affine.model,
                                          ew_constr.affine.linear,
                                          ew_constr.affine.const,
                                          ew_constr.sense)

            ro_constr.append(ew_constr)

        return ro_constr

    def dro_to_roc(self, constr):

        drule_list = self.rule_var()
        num_var = self.vt_model.vars[-1].last
        num_scen = self.num_scen
        num_rand = self.sup_model.vars[-1].last

        # Ambiguity set of the constr
        ambset = constr.ambset if constr.ambset else self.obj_ambiguity
        if not ambset:
            raise ValueError('The ambiguity set is undefined.')
        mixed_support = ambset.mix_support(primal=False)
        p = ambset.mix_model.vars[0][:num_scen]
        var_exp_list = ambset.mix_model.vars[1:]
        num_event = len(ambset.exp_constr)

        if isinstance(constr, ExpPWConstr):
            linears = []
            consts = []
            raffines = []
            for piece in constr.pieces:
                if isinstance(piece, DecLinConstr):
                    linears.append(piece.linear)
                    const = - piece.const
                    consts.append(const.reshape((const.size, 1)))
                    raffines.append(None)
                else:
                    linears.append(piece.affine.linear)
                    const = piece.affine.const
                    consts.append(const.reshape((const.size, 1)))
                    raffines.append(piece.raffine)
        elif isinstance(constr, DecLinConstr):
            linears = [constr.linear]
            const = - constr.const
            consts = [const.reshape((const.size, 1))]
            raffines = [None]
        else:
            linears = [constr.affine.linear]
            const = constr.affine.const
            consts = [const.reshape([const.size, 1])]
            raffines = [constr.raffine]

        ro_constr = []
        for i in range(linears[0].shape[0]):
            alpha = self.ro_model.dvar(num_scen)
            if num_event:
                beta = self.ro_model.dvar((num_rand, num_event))
            else:
                beta = 0

            left = alpha @ p
            for j in range(num_event):
                left += var_exp_list[j][:num_rand] @ beta[:, j]
            ro_constr.extend((left <= 0).le_to_rc(mixed_support))

            z = Vars(self.sup_model, 0, (num_rand,), 'C', None)
            for s in range(num_scen):
                drule = drule_list[s]
                for linear, raffine, const in zip(linears, raffines, consts):
                    left = linear[i, :num_var] @ drule + const[i]
                    if raffine is not None:
                        if isinstance(drule, RoAffine):
                            extra = left.raffine
                            temp = drule.affine
                            left = left.affine.reshape(left.shape)
                        elif isinstance(drule, Affine):
                            extra = 0
                            temp = drule
                        else:
                            raise TypeError('Incorrect data type.')
                        row_ind = i*num_rand + np.arange(num_rand, dtype=int)
                        new_raffine = raffine.linear[row_ind] @ temp
                        new_raffine = new_raffine.reshape((1, new_raffine.size))
                        new_raffine += raffine.const[i, :num_rand] + extra
                        left = RoAffine(new_raffine, left, self.sup_model)

                    event_indices = [k for k in range(num_event)
                                     if s in ambset.exp_constr_indices[k]]
                    if len(event_indices) > 0:
                        right = alpha[s] + (z @ beta[:, event_indices]).sum()
                    else:
                        right = alpha[s]
                    inequality = (left <= right)
                    if isinstance(inequality, RoConstr):
                        ro_constr.append(inequality.forall(ambset.sup_constr[s]))
                    elif isinstance(inequality, LinConstr):
                        ro_constr.append(inequality)
                    else:
                        raise TypeError('Incorrect data type.')

        return ro_constr

    def solve(self, solver=None, display=True, params={}):
        """
        Solve the model with the selected solver interface.

        Parameters
        ----------
            solver : {None, lpg_solver, clp_solver, ort_solver, eco_solver
                      cpx_solver, grb_solver, msk_solver, cpt_solver}
                Solver interface used for model solution. Use default solver
                if solver=None.
            display : bool
                Display option of the solver interface.
            params : dict
                A dictionary that specifies parameters of the selected solver.
                So far the argument only applies to Gurobi, CPLEX, and Mosek.
        """

        if solver is None:
            solution = def_sol(self.do_math(), display, params)
        else:
            solution = solver.solve(self.do_math(), display, params)

        if isinstance(solution, Solution):
            self.ro_model.solution = solution
        else:
            self.ro_model.solution = None

        self.ro_model.rc_model.solution = solution
        self.solution = solution

    def get(self):
        """
        Return the optimal objective value of the solved model.

        Notes
        -----
        An error message is given if the model is unsolved or no solution
        is obtained.
        """

        if self.solution is None:
            raise RuntimeError('The model is unsolved or no solution is obtained')

        solution = self.solution
        if np.isnan(solution.objval):
            msg = 'No solution available. '
            msg += f'{solution.solver} solution status: {solution.status}'
            raise RuntimeError(msg)

        return self.sign * self.solution.objval

    def optimal(self):

        return self.solution is not None


class Ambiguity:
    """
    The Ambiguity class creates an ambiguity set object
    """

    def __init__(self, model):

        self.model = model

        self.sup_constr = [None] * model.num_scen
        self.exp_constr = []
        self.exp_constr_indices = []
        self.mix_model = None

        p = self.model.p
        self.pro_constr = [p >= 0, sum(p) == 1]
        self.s = Scen(self, self.model.series_scen, self.model.p)

        self.update = True

    def __str__(self):

        return self.s.__str__()

    def __repr__(self):

        return self.s.__repr__()

    def showevents(self):
        """
        Return a data frame showing how the event-wise support sets and
        uncertainty sets of random variable expectations are defined.
        """

        num_scen = self.model.num_scen
        table = pd.DataFrame([False]*num_scen,
                             columns=['support'],
                             index=self.s.series.index)
        sup_constr = self.sup_constr
        defined = pd.notnull(pd.Series(sup_constr,
                                       index=self.s.series.index))
        table.loc[defined, 'support'] = True

        exp_constr_indices = self.exp_constr_indices
        count = 0
        for indices in exp_constr_indices:
            column = 'expectation {0}'.format(count)
            count += 1
            table[column] = False
            table.iloc[indices, count] = True

        return table

    def __getitem__(self, indices):

        return self.s[indices]

    @property
    def loc(self):

        return self.s.loc

    @property
    def iloc(self):

        return self.s.iloc

    def suppset(self, *args):
        """
        Specify the support set(s) of an ambiguity set.

        Parameters
        ----------
        args : Constraints or iterables
            Constraints or collections of constraints as iterable type of
            objects, used for defining the feasible region of the support set.

        Notes
        -----
        RSOME leaves the support set unspecified if the given argument is
        an empty iterable object.
        """

        self.update = True
        return self.s.suppset(*args)

    def exptset(self, *args):
        """
        Specify the uncertainty set of the expected values of random
        variables for an ambiguity set.

        Parameters
        ----------
        args : tuple
            Constraints or collections of constraints as iterable type of
            objects, used for defining the feasible region of the uncertainty
            set of expectations.

        Notes
        -----
        RSOME leaves the uncertainty set of expectations unspecified if the
        input argument is an empty iterable object.
        """

        self.update = True
        return self.s.exptset(*args)

    def probset(self, *args):
        """
        Specify the uncertainty set of the scenario probabilities for an
        ambiguity set.

        Parameters
        ----------
        args : tuple
            Constraints or collections of constraints as iterable type of
            objects, used for defining the feasible region of the uncertainty
            set of scenario probabilities.

        Notes
        -----
        RSOME leaves the uncertainty set of probabilities unspecified if the
        input argument is an empty iterable object.
        """

        self.update = True
        for arg in args:
            if arg.model is not self.model.pro_model:
                raise ValueError('Constraints are not defined for the ' +
                                 'probability set.')

        pr = self.model.p
        self.pro_constr = [pr >= 0, pr.sum() == 1] + list(args)

    def mix_support(self, primal=True):

        if not self.update and self.mix_model is not None:
            return self.mix_model.do_math(primal, obj=False)

        self.model.pro_model.reset()
        self.model.pro_model.st(self.pro_constr)
        pro_support = self.model.pro_model.do_math(obj=False)
        self.mix_model = GCPModel(nobj=True, mtype='M', top=self.model)

        # Constraints for probabilities
        p = self.mix_model.dvar(pro_support.linear.shape[1])
        constr = LinConstr(self.mix_model,
                           pro_support.linear, pro_support.const,
                           pro_support.sense)
        self.mix_model.st(constr)
        for q in pro_support.qmat:
            cconstr = ConeConstr(self.mix_model, p, q[1:], p, q[0])
            self.mix_model.st(cconstr)
        for ex in pro_support.xmat:
            econstr = ExpConstr(self.mix_model, p[ex[0]], p[ex[1]], p[ex[2]])
            self.mix_model.st(econstr)

        # Constraints for expectations
        for econstr, indices in zip(self.exp_constr, self.exp_constr_indices):
            self.model.exp_model.reset()
            self.model.exp_model.st(econstr)
            exp_support = self.model.exp_model.do_math(obj=False)
            exp_var = self.mix_model.dvar(exp_support.linear.shape[1])
            affine = (exp_support.linear @ exp_var -
                      p[indices].sum() * exp_support.const)
            constr = LinConstr(affine.model, affine.linear, affine.const,
                               exp_support.sense)
            self.mix_model.st(constr)
            for q in exp_support.qmat:
                cconstr = ConeConstr(self.mix_model, exp_var, q[1:],
                                     exp_var, q[0])
                self.mix_model.st(cconstr)

        return self.mix_model.do_math(primal, obj=False)
